using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using SpringPCSCAPI;
using SpringProxAPI;
using System.Runtime.InteropServices;
using System.Threading;
using System.Data.SQLite;

namespace NFCTag
{
    public partial class TagFrm : Form
    {
        public struct TagAPDUSt
        {
            public byte CLA;
            public byte INS;
            public byte P1;
            public byte P2;
            public byte Lc;
            public byte[] Data;
            public byte Le;
        }
        public TagAPDUSt TagAPDU;

        public string StatusStr = "";

        public TagFrm()
        {
            TagAPDU = new TagAPDUSt();
            TagAPDU.Data = new byte[SPCSC.MAX_RECV_LEN + 1];

            ThAPDUDeleg = new ThreadFinishedDelegate(onAPDUDone);

            InitializeComponent();            
        }
        /************************************/

        #region ArrayToHex(...)
        private string ArrayToHex(byte[] array, int len, bool reverse)
        {
            string str = "";
            int i;

            if (reverse)
            {
                for (i = len - 1; i >= 0; i--)
                    str += String.Format("{0:X02}", array[i]);
            }
            else
            {
                for (i = 0; i < len; i++)
                    str += String.Format("{0:X02}", array[i]);
            }

            return str;
        }
        #endregion

        #region h2q(...)
        byte h2q(char h)
        {
            if ((h >= '0') && (h <= '9')) return (byte)(h - '0');
            if ((h >= 'A') && (h <= 'F')) return (byte)(h - 'A' + 10);
            if ((h >= 'a') && (h <= 'f')) return (byte)(h - 'a' + 10);
            return 0x00;
        }
        #endregion

        #region h2b(...)
        byte h2b(char h1, char h2)
        {
            return (byte)((h2q(h1) << 4) + h2q(h2));
        }
        #endregion

        #region HexToArray(...)
        bool HexToArray(byte[] array, string str, int size, bool reverse)
        {
            int i, l;
            if ((str == null) || (array == null))
                return false;

            l = str.Length / 2;
            for (i = 0; i < size; i++)
            {
                if (reverse)
                {
                    if (i < l)
                        array[size - i - 1] = h2b(str[2 * i], str[(2 * i) + 1]);
                    else
                        array[size - i - 1] = 0x00;
                }
                else
                {
                    if (i < l)
                        array[i] = h2b(str[2 * i], str[(2 * i) + 1]);
                    else
                        array[i] = 0x00;
                }
            }
            return true;
        }
        #endregion

        /************************************/

        #region Print APDU
        public void printAPDURcv()
        {
            apduRcvTb.Text  = String.Format("{0:X02}", TagAPDU.CLA);
            apduRcvTb.Text += String.Format("{0:X02}", TagAPDU.INS);
            apduRcvTb.Text += String.Format("{0:X02}", TagAPDU.P1);
            apduRcvTb.Text += String.Format("{0:X02}", TagAPDU.P2);
            apduRcvTb.Text += String.Format("{0:X02}", TagAPDU.Lc);
            apduRcvTb.Text += ArrayToHex(TagAPDU.Data, TagAPDU.Lc, false);
            apduRcvTb.Text += String.Format("{0:X02}", TagAPDU.Le);        
        }
        #endregion

        /************************************/

        #region Reader

        string readerStr = "";

        public void SetReader(string readerName)
        {
            SPCSC.SetReader(readerName);
            readerStr = readerName;
        }

        public void ConnectReader()
        {
            uint rc = SPCSC.directConnect();
            if (rc != SPCSC.SPCSC_SUCCESS)
            {
                statusTb.Text = "Connect Error : " + rc.ToString() + " " + SPCSC.codeToStr(rc);
                return;
            }

            StopSlot();
            GetNFCID1();
            GetNFCID2();
            GetNFCID3();
            GetATS();
            StartTarget();

            timerStatus.Enabled = true;

            StatusStr = "Tag Ready";
        }

        public bool StopSlot()
        {
            APDURsp apduOut = new APDURsp();
            byte[] apduIn = { 0x58, 0x22 };
            try
            {
                uint rc = SPCSC.Control(apduIn, ref apduOut);
                if (rc != SPCSC.SPCSC_SUCCESS)
                {
                    StatusStr = "Slot Stop Error : " + rc.ToString() + " " + SPCSC.codeToStr(rc);
                    return false;
                }
            }
            catch (Exception e)
            {
                StatusStr = "Slot Stop Error : " + e.Message;  
                return false; 
            }
            return true;
        }

        private void TagFrm_Shown(object sender, EventArgs e)
        {
            /* Showing configuration form - User has to select reader */
            ChooseReaderFrm cf = new ChooseReaderFrm(this, true);
            cf.ReadRegister();

            if ((cf.ReaderName == "") || (!cf.connect))
                cf.ShowDialog();
            else
                SetReader(cf.ReaderName);

            ConnectReader();
        }

        /* Display ChooseReaderFrm to select another reader */
        private void ChangeReaderBt_Click(object sender, EventArgs e)
        {
            timerStatus.Enabled = false;

            SPCSC.Disconnect();

            ChooseReaderFrm cf = new ChooseReaderFrm(this, false);
            cf.ShowDialog();

            ConnectReader();
        }

        #endregion

        /************************************/

        #region Thread
        /* check every 100ms get APDU */
        System.Threading.Thread APDUThread = null;
        public ThreadFinishedDelegate ThAPDUDeleg;
        public delegate void ThreadFinishedDelegate();
        
        public void ThAPDUProc()
        {
            if (!GetAPDU())
                return;

            try
            {
                Invoke(ThAPDUDeleg);
            }
            catch (Exception e) { }
        }

        private void timerStatus_Tick(object sender, EventArgs e)
        {
            
            APDUThread = new System.Threading.Thread(new System.Threading.ThreadStart(ThAPDUProc));
            APDUThread.Start();

            statusTb.Text = StatusStr;
            printAPDURcv();
        }

        public void onAPDUDone()
        {
            
        }
        #endregion

        /************************************/

        #region NFC Configuration
        public bool GetNFCID1()
        {
            APDURsp apduOut = new APDURsp();
            byte[] apduIn = { 0x85, 0x00 };
            uint rc = SPCSC.Control(apduIn, ref apduOut);
            if (rc == SPCSC.SPCSC_SUCCESS)
            {
                if (apduOut.APDUArray[0] != 0x00)
                {
                    short error = (short)(0 - apduOut.APDUArray[0]);
                    StatusStr = "NFCID1 Error : " + error.ToString() + " " + SPROX.GetErrorMessage(error);
                    return false;
                }

                atqaTb.Text = String.Format("{0:X02}{1:X02}", apduOut.APDUArray[1], apduOut.APDUArray[2]);
                nfcid1Tb.Text = "";
                for (int i = 3; i < 6; i++)
                    nfcid1Tb.Text += String.Format("{0:X02}", apduOut.APDUArray[i]);
                sakTb.Text = String.Format("{0:X02}", apduOut.APDUArray[6]);
            }
            else
            {
                StatusStr = "NFCID1 Error : " + rc.ToString() + " " + SPCSC.codeToStr(rc);
                return false;
            }

            return true;
        }

        public bool SetNFCID1()
        {
            if (atqaTb.Text.Length != 4)
            {
                StatusStr = "NFCID1 Error : Length ATQA incorrect";
                return false;
            }
            if (nfcid1Tb.Text.Length != 6)
            {
                StatusStr = "NFCID1 Error : Length NFCID1 incorrect";
                return false;
            }
            if (sakTb.Text.Length != 2 )
            {
                StatusStr = "NFCID1 Error : Length SAK incorrect";
                return false;
            }

            string str = "8500" + atqaTb.Text + nfcid1Tb.Text + sakTb.Text;
            
            /* First we test if the text entered is hexadecimal */
            Regex regHex = new Regex("^[0-9a-fA-F]*$");
            if (!regHex.IsMatch(str))
            {
                StatusStr = "NFCID1 Error : Invalid format. Only hexadecimal allowed\n";
                return false;
            }

            byte[] apduIn = new byte[str.Length / 2];
            HexToArray(apduIn, str, apduIn.Length, false);
            APDURsp apduOut = new APDURsp();
            uint rc = SPCSC.Control(apduIn, ref apduOut);
            if (rc == SPCSC.SPCSC_SUCCESS)
            {
                if (apduOut.APDUArray[0] != 0x00)
                {
                    short error = (short)(0 - apduOut.APDUArray[0]);
                    StatusStr = "NFCID1 Error : " + error.ToString() + " " + SPROX.GetErrorMessage(error);
                    return false;
                }
            }
            else
            {
                StatusStr = "NFCID1 Error : " + rc.ToString() + " " + SPCSC.codeToStr((uint)rc);
                return false;
            }

            return true;
        }

        public bool GetNFCID2()
        {
            APDURsp apduOut = new APDURsp();
            byte[] apduIn = { 0x85, 0x01 };
            uint rc = SPCSC.Control(apduIn, ref apduOut);
            if (rc == SPCSC.SPCSC_SUCCESS)
            {
                if (apduOut.APDUArray[0] != 0x00)
                {
                    short error = (short)(0 - apduOut.APDUArray[0]);
                    StatusStr = "NFCID2 Error : " + error.ToString() + " " + SPROX.GetErrorMessage(error);
                    return false;
                }

                nfcid2Tb.Text = "";
                for (int i = 1; i < 7; i++)
                    nfcid2Tb.Text += String.Format("{0:X02}", apduOut.APDUArray[i]);
                padTb.Text = "";
                for (int i = 7; i < 15; i++)
                    padTb.Text += String.Format("{0:X02}", apduOut.APDUArray[i]);
                sysCodeTb.Text = String.Format("{0:X02}{1:X02}", apduOut.APDUArray[15], apduOut.APDUArray[16]);
            }
            else
            {
                StatusStr = "NFCID2 Error : " + rc.ToString() + " " + SPCSC.codeToStr((uint)rc);
                return false;
            }

            return true;
        }

        public bool SetNFCID2()
        {
            if (nfcid2Tb.Text.Length != 12)
            {
                StatusStr = "NFCID2 Error : Length NFCID2 incorrect";
                return false;
            }
            if (padTb.Text.Length != 16)
            {
                StatusStr = "NFCID2 Error : Length PAD incorrect";
                return false;
            }
            if (sysCodeTb.Text.Length != 4)
            {
                StatusStr = "NFCID2 Error : Length System Code incorrect";
                return false;
            }

            string str = "8501" + nfcid2Tb.Text + padTb.Text + sysCodeTb.Text;

            /* First we test if the text entered is hexadecimal */
            Regex regHex = new Regex("^[0-9a-fA-F]*$");
            if (!regHex.IsMatch(str))
            {
                StatusStr = "NFCID2 Error : Invalid format. Only hexadecimal allowed\n";
                return false;
            }

            byte[] apduIn = new byte[str.Length / 2];
            HexToArray(apduIn, str, apduIn.Length, false);
            APDURsp apduOut = new APDURsp();
            uint rc = SPCSC.Control(apduIn, ref apduOut);
            if (rc == SPCSC.SPCSC_SUCCESS)
            {
                if (apduOut.APDUArray[0] != 0x00)
                {
                    short error = (short)(0 - apduOut.APDUArray[0]);
                    StatusStr = "NFCID2 Error : " + error.ToString() + " " + SPROX.GetErrorMessage(error);
                    return false;
                }
            }
            else
            {
                StatusStr = "NFCID2 Error : " + rc.ToString() + " " + SPCSC.codeToStr((uint)rc);
                return false;
            }

            return true;
        }

        public bool GetNFCID3()
        {
            APDURsp apduOut = new APDURsp();
            byte[] apduIn = { 0x85, 0x02 };
            uint rc = SPCSC.Control(apduIn, ref apduOut);
            if (rc == SPCSC.SPCSC_SUCCESS)
            {
                if (apduOut.APDUArray[0] != 0x00)
                {
                    short error = (short)(0 - apduOut.APDUArray[0]);
                    StatusStr = "NFCID3 Error : " + error.ToString() + " " + SPROX.GetErrorMessage(error);
                    return false;
                }

                nfcid3Tb.Text = String.Format("{0:X02}", apduOut.APDUArray[1]);
            }
            else
            {
                StatusStr = "NFCID3 Error : " + rc.ToString() + " " + SPCSC.codeToStr((uint)rc);
                return false;
            }

            return true;
        }

        public bool SetNFCID3()
        {
            if (nfcid3Tb.Text.Length != 2)
            {
                StatusStr = "NFCID3 Error : Length NFCID3 incorrect";
                return false;
            }

            string str = "8502" + nfcid3Tb.Text;

            /* First we test if the text entered is hexadecimal */
            Regex regHex = new Regex("^[0-9a-fA-F]*$");
            if (!regHex.IsMatch(str))
            {
                StatusStr = "NFCID3 Error : Invalid format. Only hexadecimal allowed\n";
                return false;
            }

            byte[] apduIn = new byte[str.Length / 2];
            HexToArray(apduIn, str, apduIn.Length, false);
            APDURsp apduOut = new APDURsp();
            uint rc = SPCSC.Control(apduIn, ref apduOut);
            if (rc == SPCSC.SPCSC_SUCCESS)
            {
                if (apduOut.APDUArray[0] != 0x00)
                {
                    short error = (short)(0 - apduOut.APDUArray[0]);
                    StatusStr = "NFCID3 Error : " + error.ToString() + " " + SPROX.GetErrorMessage(error);
                    return false;
                }
            }
            else
            {
                StatusStr = "NFCID3 Error : " + rc.ToString() + " " + SPCSC.codeToStr((uint)rc);
                return false;
            }

            return true;
        }

        public bool GetATS()
        {
            APDURsp apduOut = new APDURsp();
            byte[] apduIn = { 0x85, 0x03 };
            uint rc = SPCSC.Control(apduIn, ref apduOut);
            if (rc == SPCSC.SPCSC_SUCCESS)
            {
                if (apduOut.APDUArray[0] != 0x00)
                {
                    short error = (short)(0 - apduOut.APDUArray[0]);
                    StatusStr = "ATS Error : " + error.ToString() + " " + SPROX.GetErrorMessage(error);
                    return false;
                }
                
                atsTb.Text = "";
                for (int i = 1; i <= apduOut.APDUArray[1]; i++)
                    atsTb.Text += String.Format("{0:X02}", apduOut.APDUArray[i]);
            }
            else
            {
                StatusStr = "ATS Error : " + rc.ToString() + " " + SPCSC.codeToStr((uint)rc);
                return false;
            }

            return true;
        }

        public bool SetATS()
        {
            string str = "8503" + atsTb.Text;

            /* First we test if the text entered is hexadecimal */
            Regex regHex = new Regex("^[0-9a-fA-F]*$");
            if (!regHex.IsMatch(str))
            {
                StatusStr = "ATS Error : Invalid format. Only hexadecimal allowed\n";
                return false;
            }

            byte[] apduIn = new byte[str.Length / 2];
            HexToArray(apduIn, str, apduIn.Length, false);
            APDURsp apduOut = new APDURsp();
            uint rc = SPCSC.Control(apduIn, ref apduOut);
            if (rc == SPCSC.SPCSC_SUCCESS)
            {
                if (apduOut.APDUArray[0] != 0x00)
                {
                    short error = (short) (0 - apduOut.APDUArray[0]);
                    StatusStr = "ATS Error : " + error.ToString() + " " + SPROX.GetErrorMessage(error);
                    return false;
                }
            }
            else
            {
                StatusStr = "ATS Error : " + rc.ToString() + " " + SPCSC.codeToStr((uint)rc);
                return false;
            }

            return true;
        }

        private void ReadTagCfgBt_Click(object sender, EventArgs e)
        {
            if (!GetNFCID1())
                return;
            if (!GetNFCID2())
                return;
            if (!GetNFCID3())
                return;
            if (!GetATS())
                return;
            if (!StartTarget())
                return;

            StatusStr = "Tag Ready";
        }

        private void WriteTagCfgBt_Click(object sender, EventArgs e)
        {
            if (!SetNFCID1())
                return;
            if (!SetNFCID2())
                return;
            if (!SetNFCID3())
                return;
            if (!SetATS())
                return;
            if (!StartTarget())
                return;

            StatusStr = "Tag Ready";
        }

        #endregion

        /************************************/

        #region NFC Function
        public bool StartTarget()
        {
            APDURsp apduOut = new APDURsp();
            byte[] apduIn = { 0x83, 0x01 };

            try
            {
                uint rc = SPCSC.Control(apduIn, ref apduOut);
                if (rc != SPCSC.SPCSC_SUCCESS)
                {
                    StatusStr = "Start Target Error : " + rc.ToString() + " " + SPCSC.codeToStr((uint)rc);
                    return false;
                }

                if (apduOut.APDUArray[0] != 0x00)
                {
                    short error = (short)(0 - apduOut.APDUArray[0]);
                    StatusStr = "Start Target Error : " + error.ToString() + " " + SPROX.GetErrorMessage(error);
                    return false;
                }
            }
            catch (Exception e)
            {
                StatusStr = "Start Target Error : " + e.Message;
                return false;
            }

            return true;
        }

        private void StartBt_Click(object sender, EventArgs e)
        {
            if (!StartTarget())
                return;

            StatusStr = "Tag Ready";
        }

        public bool GetAPDU()
        {
            APDURsp apduOut = new APDURsp();
            byte[] apduIn = { 0x84 };

            try
            {
                uint rc = SPCSC.Control(apduIn, ref apduOut);
                if (rc != SPCSC.SPCSC_SUCCESS)
                {
                    StatusStr = "Get APDU Error : " + rc.ToString() + " " + SPCSC.codeToStr(rc);
                    return false;
                }


                if (apduOut.APDUArray[0] != 0x00)
                {
                    if (apduOut.APDUArray[0] != 0x03)
                    {
                        short error = (short)(0 - apduOut.APDUArray[0]);
                        StatusStr = "Get APDU Error : " + error.ToString() + " " + SPROX.GetErrorMessage(error);
                    }
                    return false;
                }

                TagAPDU.CLA = apduOut.APDUArray[1];
                TagAPDU.INS = apduOut.APDUArray[2];
                TagAPDU.P1 = apduOut.APDUArray[3];
                TagAPDU.P2 = apduOut.APDUArray[4];
                TagAPDU.Lc = apduOut.APDUArray[5];
                for (int i = 0; i < TagAPDU.Lc; i++)
                    TagAPDU.Data[i] = apduOut.APDUArray[6 + i];
                TagAPDU.Lc = apduOut.APDUArray[6 + TagAPDU.Lc];

            }
            catch (Exception e)
            {
                StatusStr = "Get APDU Error : " + e.Message;
                return false;
            }

            StatusStr = "APDU received";
            return true;
        }

        public bool SetAPDU(string apduStr)
        {
            string str = "84" + apduStr;

            /* First we test if the text entered is hexadecimal */
            Regex regHex = new Regex("^[0-9a-fA-F]*$");
            if (!regHex.IsMatch(str))
            {
                StatusStr = "Set APDU Error : Invalid format. Only hexadecimal allowed\n";
                return false;
            }

            byte[] apduIn = new byte[str.Length / 2];
            HexToArray(apduIn, str, apduIn.Length, false);
            APDURsp apduOut = new APDURsp();
            uint rc = SPCSC.Control(apduIn, ref apduOut);
            if (rc == SPCSC.SPCSC_SUCCESS)
            {
                if (apduOut.APDUArray[0] != 0x00)
                {
                    short error = (short)(0 - apduOut.APDUArray[0]);
                    StatusStr = "Set APDU Error : " + error.ToString() + " " + SPROX.GetErrorMessage(error);
                    return false;
                }
            }
            else
            {
                StatusStr = "Set APDU Error : " + rc.ToString() + " " + SPCSC.codeToStr(rc);
                return false;
            }

            StatusStr = "APDU sent";

            return true;
        }

        private void apduSndBt_Click(object sender, EventArgs e)
        {
            SetAPDU(apduSndTb.Text);
        }

        #endregion

        /************************************/

        #region SQLite

        private DataSet sqlDataSet = new DataSet();
        private DataTable sqlDataTable = new DataTable();

        public void CreateDesfire()
        {            
            sqlConnection.Open();
            sqlDataSet.Reset();
            sqlDataAdaptater.Fill(sqlDataSet);
            sqlDataTable = sqlDataSet.Tables[0];
            dataGridView1.DataSource = sqlDataTable;
            sqlConnection.Close();
        }
        #endregion
        /************************************/

        private void TagFrm_FormClosing(object sender, FormClosingEventArgs e)
        {
            SPCSC.Disconnect();
        }

        private void SelectBt_Click(object sender, EventArgs e)
        {
            CreateDesfire();
        }
    }
}