using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using SpringPCSCAPI;
using System.Runtime.InteropServices;
using System.Threading;

namespace NFCTool
{
  public partial class PosterFrm : Form
  {
    string Card_URI_prefix = "";
    int TYPE_SP = 0x53;
    int TYPE_T = 0x54;
    int TYPE_U = 0x55;
    string Card_URI_content = "";
    string Card_URI_Tot = "";
    string Card_Title_lang = "";
    string Card_Title_content = "";
    bool Card_Title_present = false;
    int type_NDEF = 0;
    private static Mutex mut = new Mutex(false, @"Global\SpringCard.NFCApp");

    public PosterFrm()
    {
      // for card status 
      ThStatusDeleg = new ThreadFinishedDelegate(onStatusDone);
      InitializeComponent();
      set_URI_Codes();
      setTooltips();
      LoadMifareKeys();
     // SetReader(reader);
    }

    /* This function is called as soon as a card is detected */
    private void processCard()
    {
      ClearTabChoices();
      GetCardType();
      ReadCard();
      decodeTLV(card_content);
      type_NDEF = 0;
      decodeNDEF(V_VALUE);
      displayResults();
    }


#region graphics management
    private void setTooltips()
    {
      ToolTip WriteTp = new ToolTip();
      WriteTp.SetToolTip(butWrite, "Write NDEF on card");
      ToolTip CopyTp = new ToolTip();
      CopyTp.SetToolTip(butCopy, "Copy NDEF to ClipBoard");
    }

    private void ClearTabChoices()
    {
      rTBURIContent.Clear();
      rTBText.Clear();
      tBtitleSp.Clear();
      rTBURISp.Clear();
      tBtitleSp.Clear();
      cBTitle.Checked = false;
      Card_URI_prefix = "";
      Card_URI_content = "";
      Card_URI_Tot = "";
      Card_Title_lang = "";
      Card_Title_content = "";
      Card_Title_present = false;
      type_NDEF = 0;

    }

    private void ReaderEmpty()
    {
      rTBAdvise.Visible = true;
      rTBAdvise.Text = "Waiting for a card";
      butWrite.Visible = false;
      Card_Title_present = false;
      lblSize.Text = "";
    }

    private void cBTitle_CheckedChanged(object sender, EventArgs e)
    {
      tBtitleSp.Enabled = cBTitle.Checked;
      cBLangTitleSp.Enabled = cBTitle.Checked;
    }

    private void pBNew_Click(object sender, EventArgs e)
    {
      rTBAdvise.Visible = false;

      labelTitle.Visible = true;
      cBTitle.Visible = true;
      labelUri.Visible = true;
      butWrite.Visible = true;
      labelContent.Visible = true;
      labelPrefix.Visible = true;
      comboPrefixSp.Visible = true;
    }

    private void displayResults()
    {
      if (type_NDEF == TYPE_SP)
      {
        comboPrefixSp.Text = Card_URI_prefix;
        rTBURISp.Text = Card_URI_content;
        if (Card_Title_present)
        {
          cBTitle.Checked = true;
          cBLangTitleSp.Text = Card_Title_lang;
          tBtitleSp.Text = Card_Title_content;
        }
      }
      if (type_NDEF == TYPE_U)
      {
        cBPrefixURI.Text = Card_URI_prefix;
        rTBURIContent.Text = Card_URI_content;
      }
      if (type_NDEF == TYPE_T)
      {
        cBLangText.Text = Card_Title_lang;
        rTBText.Text = Card_Title_content;
      }
      rTBNDEFEnc.Text = createStrFromByteArray(V_VALUE);
      rTBAdvise.Visible = true;
      rTBAdvise.Text = "";
      if (card_type == ULTRALIGHT)
        rTBAdvise.Text = "UltraLight";
      if(card_type == ULTRALIGHTC)
        rTBAdvise.Text = "UltraLight C";
      if (card_type == JEWEL)
        rTBAdvise.Text = "Jewel";
      if (card_type == MIFARE_1K || card_type == MIFARE_4K)
        rTBAdvise.Text = "Mifare Std";
      rTBAdvise.Text += " Card found";
      if(Card_URI_Tot.StartsWith("http"))
        RunUrl(Card_URI_Tot,true);
    }

#endregion

#region reader

    string readerStr = "";
    /* private void Form1_Shown(object sender, EventArgs e)
     {
       // Showing configuration form - User has to select reader 
       ChooseReaderFrm cf = new ChooseReaderFrm(this, true);
       cf.ReadRegister();

       if ((cf.ReaderName == "") || (!cf.connect))
         cf.ShowDialog();
       else
         SetReader(cf.ReaderName);
     }*/

    public void SetReader(string readerName)
    {
      SPCSC.SetReader(readerName);
      readerStr = readerName;
      timerStatus.Enabled = true;
    }
#endregion

#region conversion from str to byte and vice versa

    private byte[] createByteArrayFromStr(string StrIn)
    {
      string Str = StrIn.Replace("0x", "").Replace(" ", "");

      if (!Str.Contains(" "))
      {
        string nStr = "";
        for (int cpt = 0; cpt < Str.Length; cpt = cpt + 2)
        {
          nStr = nStr + Str.Substring(cpt, 2) + " ";
        }
        Str = nStr;
      }

      // command stored into byte array
      char[] delimiterByte = new char[1];
      delimiterByte[0] = ' ';
      string properStr = Str;
      // removing last " "
      if (Str.EndsWith(" "))
      {
        properStr = Str.Substring(0, Str.Length - 1);
      }

      string[] Strbuff = properStr.Split(delimiterByte);
      byte[] StrBuffer = new byte[Strbuff.Length];
      for (int i = 0; i < Strbuff.Length; i++)
      {
        // converting byte string to byte  
        StrBuffer[i] = System.Convert.ToByte(int.Parse(Strbuff[i],
          System.Globalization.NumberStyles.HexNumber));
      }
      return StrBuffer;
    }

    private string createStrFromByteArray(byte[] byteArray)
    {
      string Str = "";
      if (byteArray.Length <= 0)
        return Str;
      for (int cpt = 0; cpt < byteArray.Length - 1; cpt++)
        Str += String.Format("{0:X02}", byteArray[cpt]) + " ";
      Str += String.Format("{0:X02}", byteArray[byteArray.Length - 1]);
      return Str;
    }

    private byte[] createByteArrayFromAsciiStr(string StrIn)
    {
      byte[] array = new byte[StrIn.Length];

      for (int cpt = 0; cpt < StrIn.Length; cpt++)
        array[cpt] = (byte)System.Convert.ToChar(StrIn[cpt]);

      return array;

    }

    private byte StringToByte(string StrIn)
    {
      string Str = StrIn.Replace("0x", "").Replace(" ", "");
      return System.Convert.ToByte(int.Parse(Str,
          System.Globalization.NumberStyles.HexNumber));
    }

    private string createAsciiStrFromByteArray(byte[] byteArray)
    {
      string Str = "";
      if (byteArray.Length <= 0)
        return Str;
      for (int cpt = 0; cpt < byteArray.Length; cpt++)
        Str += byteToAscii(byteArray[cpt]);
      return Str;
    }

    public string byteToAscii(byte value)
    {
      if (value < 32 || value > 127)
        return (".");
      else
        return (System.Convert.ToChar(value).ToString());
    }

    public string byteToString(byte value)
    {
      if (value < 32 || value > 127)
        return (String.Format("{0:X02}", value));
      else
        return (System.Convert.ToChar(value).ToString());
    }

    public string byteToStringHex(byte value)
    {
      return ("0x" + String.Format("{0:X02}", value));
    }
#endregion

#region determine card type
    int pixnnIndex = 26;
    const int JEWEL = 1;
    const int ULTRALIGHT = 2;
    const int ULTRALIGHTC = 5;
    const int DESFIRE = 4;
    const int MIFARE_1K = 6;
    const int MIFARE_4K = 7;
    int card_type;

    private void GetCardType()
    {
      string AtrStr = SPCSC.GetAtrStr();
      if (AtrStr.Substring(pixnnIndex, 4) == "0003")
        card_type = ULTRALIGHT;
      if (AtrStr.Substring(pixnnIndex, 4) == "0030")
        card_type = JEWEL;
      if (AtrStr.Substring(pixnnIndex, 4) == "0001")
        card_type = MIFARE_1K;
      if (AtrStr.Substring(pixnnIndex, 4) == "0002")
        card_type = MIFARE_4K;
    }
#endregion

#region read card

    const int MAX_NDEF_UL = 46;
    const int MAX_NDEF_ULC = 142;
    const int MAX_NDEF_JEWEL = 90;

    byte[] card_content;

    const long NDEF_TAG_APPLICATION_ID = 0xD2760000850100;
    const uint CAPACITY_CONTAINER_FILE_ID = 0xE103;
    public byte[] DESFIRE_BLANK_KEY =
  new byte[16] { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    };
    public byte[] DESFIRE_KEY_0 =
      new byte[16] { 0xB0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x11, 0x22, 0x33
    };
    public byte[] DESFIRE_KEY_1 =
      new byte[16] { 0xA0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x11, 0x22, 0x33
    };


    private void ReadCard()
    {
      if (card_type == JEWEL)
        ReadJewel();
      if (card_type == ULTRALIGHT)
        ReadMifareUL();
      if (card_type == DESFIRE)
        ReadDEsfire();
      if (card_type == MIFARE_1K || card_type == MIFARE_4K)
        ReadMifare1K();
     
    }


    #region JEWEL
    private void ReadJewel()
    {
      APDUCmd readUl = new APDUCmd("FF B0 00 0C 6C");
      APDURsp RspRead = ConnectAndTransmit(readUl);

      card_content = RspRead.DataOut;

    }
    #endregion

    #region DESFIRE



    private bool ReadDEsfire()
    {
      /*byte[] buffer = new byte[4096];
      uint length = 0;
      int i;

      
      if (pcscSelectApp(SPCSC.GetHCard(), NDEF_TAG_APPLICATION_ID) !=0)
        return false;
      if (pcscAuth(SPCSC.GetHCard(), 1, DESFIRE_KEY_1) != 0)
        return false;
      if (pcscRead(SPCSC.GetHCard(), CAPACITY_CONTAINER_FILE_ID, 0, 0, buffer,
        ref length) != 0)
        return false;

      */
      return true;
    }

    #endregion

    #region Mifare 1K

    public void ReadMifare1K()
    {
      /* step 1 : read MAD sector */
      APDUCmd readUl = new APDUCmd("FF F3 00 01 20");
      APDURsp RspRead = ConnectAndTransmit(readUl);
     
      /* step 2 : determine the number of sectors to read */
      int nb_sectors = 0;

      for (int i = 2; i < RspRead.DataOut.Length; i=i+2)
      { 
        if( (RspRead.DataOut[i] == AID_1) && (RspRead.DataOut[i+1] == AID_2))
          nb_sectors++;
      }
      card_content = new byte[nb_sectors * SZ_SECTOR_MIF1K];
      int offset_card = 0;
      /* for each sector, get data  */
      for (int sector = 0; sector < nb_sectors; sector++)
      {
        byte block = (byte)((sector + 1) * 4 );
        APDUCmd ApduRead = new APDUCmd(0xFF, 0xF3, 0x00, block, 0x30);
        RspRead = ConnectAndTransmit(ApduRead);
        for (int cpt = 0; cpt < SZ_SECTOR_MIF1K; cpt++)
          card_content[offset_card + cpt] = RspRead.DataOut[cpt];
        offset_card += SZ_SECTOR_MIF1K;
      }
    }

    public void LoadMifareKeys()
    {
      SPCSC.ConnectCard();
      SPCSC.LoadKey(MIFARE_KEY_A, 1);
      SPCSC.LoadKey(MIFARE_KEY_NFC, 2);
      SPCSC.LoadKey(MIFARE_WRITE_KEY, 3);

      SPCSC.Disconnect();
    }

    #endregion

    #region Mifare UltraLight
    private void ReadMifareUL()
    {
      byte page;
      int offset_cardContent = 0;
      card_content = new byte[192];
      card_type = ULTRALIGHT;

      for (page = 4; page < 44; page += 4)
      {
        APDUCmd readUl = new APDUCmd(0xFF, 0xB0, 0x00, page, 0x10);
        APDURsp RspRead = ConnectAndTransmit(readUl);
        if( (RspRead.SW1 !=0x90) || (RspRead.SW2 != 0x00))
          return;
        for (int i = 0; i < 16; i++)
        {
          if ((offset_cardContent + i < card_content.Length) && (i <= RspRead.DataOut.Length)) 
            card_content[offset_cardContent + i] = RspRead.DataOut[i];
        }
        offset_cardContent += 16;
        if ((page >= 16) )
          card_type = ULTRALIGHTC;
      }

    }
    #endregion

#endregion

#region card status
    /* check every 100ms card status for selected reader, 
     * and change interface according to status found */

    bool card_present = false;

    System.Threading.Thread StatusThread = null; /* Get reader status */
    public ThreadFinishedDelegate ThStatusDeleg;
    public delegate void ThreadFinishedDelegate();
    public string StatusStr = "";

    public void ThStatusProc()
    {
      string str = SPCSC.CheckStatus();
      StatusStr = str;
      if (str.Contains(":"))
        StatusStr = StatusStr.Remove(str.IndexOf(':'));
      try
      {
        Invoke(ThStatusDeleg);
      }
      catch (Exception e) { }

    }

    private void timerStatus_Tick(object sender, EventArgs e)
    {
      StatusThread = new System.Threading.Thread(new
      System.Threading.ThreadStart(ThStatusProc));
      StatusThread.Start();
    }

    public void onStatusDone()
    {
      if ((StatusStr.Contains("PRESENT")) || (StatusStr.Contains("INUSE")))
      {
        if (!card_present)
        {
          timerStatus.Enabled = false;
          card_present = true;
          rTBAdvise.Visible = false;
          butWrite.Visible = true;
          processCard();
          timerStatus.Enabled = true;
        }
      }
      else
      {
        card_present = false;
        ReaderEmpty();
      }
    }


#endregion

#region TLV encoding/decoding

    // TLV Constants

    const int INDEX_T = 0;
    const int INDEX_L = 1;
    const byte NULL_TLV = 0x00;
    const byte LOCK_CONTROL_TLV = 0x01;
    const byte MEMORY_CONTROL_TLV = 0x02;
    const byte NDEF_MESSAGE_TLV = 0x03;
    const byte PROPRIETARY_TLV = 0xFD;
    const byte TERMINATOR_TLV = 0xFE;

    const byte FLAG_L_3_BYTES = 0xFF;

    const int MAX_L = 65534;//FFFE

    string T_STR = "";
    int L_VALUE = 0;
    byte[] V_VALUE;


    private void decodeTLV(byte[] dataIn)
    {
      if (dataIn.Length > 2)
      {
        // T
        byte t = dataIn[INDEX_T];
        T_STR = decodeT(t);

        // L
        int start_v = INDEX_L + 1;
        if (dataIn[INDEX_L] == FLAG_L_3_BYTES)
        {
          start_v += 2;
          L_VALUE = (int)(dataIn[INDEX_L + 1] << 8 + dataIn[INDEX_L + 2]);
        }
        else
        {
          L_VALUE = (int)dataIn[INDEX_L];
        }

        // V
        V_VALUE = new byte[L_VALUE];
        for (int cpt = 0; cpt < L_VALUE; cpt++)
        {
          if (dataIn.Length > cpt + start_v)
            V_VALUE[cpt] = dataIn[cpt + start_v];
        }
      }


    }


    private string decodeT(byte t)
    {
      switch (t)
      {
        case NULL_TLV: return "NULL";
        case LOCK_CONTROL_TLV: return "Lock Control";
        case MEMORY_CONTROL_TLV: return "Memory Control";
        case NDEF_MESSAGE_TLV: return "NDEF Message";
        case PROPRIETARY_TLV: return "Proprietary";
        case TERMINATOR_TLV: return "Terminator";
        default: return "Type unknown";
      }
    }

    private byte[] encodeTLV(byte type, byte[] values)
    {

      byte[] TLV_array;
      int start_v = INDEX_L + 1;

      if (values.Length < FLAG_L_3_BYTES)
      {
        TLV_array = new byte[values.Length + 2];
        TLV_array[INDEX_T] = type;
        TLV_array[INDEX_L] = (byte)values.Length;
      }
      else
      {
        int V_size;
        start_v += 2;
        V_size = values.Length;
        if (values.Length > MAX_L)
          V_size = MAX_L;
        TLV_array = new byte[V_size + 4];
        TLV_array[INDEX_T] = type;
        TLV_array[INDEX_L] = FLAG_L_3_BYTES;
        TLV_array[INDEX_L + 1] = (byte)(V_size >> 8);
        TLV_array[INDEX_L + 2] = (byte)V_size;
      }

      for (int cpt = start_v; cpt < TLV_array.Length; cpt++)
        TLV_array[cpt] = values[cpt - start_v];

      return TLV_array;

    }
#endregion

#region NDEF encoding/decoding

    const byte MESSAGE_BEGIN_FLAG = 0x80;
    const byte MESSAGE_END_FLAG = 0x40;
    const byte CHUNK_FLAG = 0x20;
    const byte SHORT_RECORD_FLAG = 0x10;
    const byte ID_LENGTH_FLAG = 0x08;

    const byte SHORT_RECORD_HEADER = 0xD1;
    const byte SR_MB_HEADER = 0x91;
    const byte SR_ME_HEADER = 0x51;
    const byte RECORD_HEADER = 0xC1;

    // TNF constants
    const byte EMPTY = 0x00;
    const byte NFC_TYPE = 0x01;
    const byte MEDIA_TYPE = 0x02;
    const byte ABSOLUTE_URI = 0x03;
    const byte EXTERNAL_TYPE = 0x04;
    const byte UNKNOWN_TYPE = 0x05;
    const byte UNCHANGED = 0x06;
    const byte RESERVED = 0x07;


    string type_ndef_str = "";
    string record_name_str = "";
    string payload_str = "";
    int payload_size;

    byte[] payload_array;

    const int TYPE_LENGTH_INDEX = 1;
    const int PAYLOAD_LENGTH_INDEX = 2;

    private string decodeTNF(byte tnf)
    {
      switch (tnf)
      {
        case EMPTY: return "Empty";
        case NFC_TYPE: return "NFC Forum well-known type";
        case MEDIA_TYPE: return "Media type as defined in RFC 2046";
        case ABSOLUTE_URI: return "Absolute URI as defined in RFC 3986";
        case EXTERNAL_TYPE: return "NFC Forum external type";
        case UNKNOWN_TYPE: return "Unknown";
        case UNCHANGED: return "Unchanged";
        case RESERVED: return "Reserved";
        default: return "Type unknown";
      }
    }

    NameValueCollection URI_Codes = new NameValueCollection();

    private void set_URI_Codes()
    {
      URI_Codes.Add("0x00", "");
      URI_Codes.Add("0x01", "http://www.");
      URI_Codes.Add("0x02", "https://www.");
      URI_Codes.Add("0x03", "http://");
      URI_Codes.Add("0x04", "https://");
      URI_Codes.Add("0x05", "tel:");
      URI_Codes.Add("0x06", "mailto:");
      URI_Codes.Add("0x07", "ftp://anonymous:anonymous@");
      URI_Codes.Add("0x08", "ftp://ftp.");
      URI_Codes.Add("0x09", "ftps://");
      URI_Codes.Add("0x0A", "sftp://");
      URI_Codes.Add("0x0B", "smb://");
      URI_Codes.Add("0x0C", "nfs://");
      URI_Codes.Add("0x0D", "ftp://");
      URI_Codes.Add("0x0E", "dav://");
      URI_Codes.Add("0x0F", "news:");
      URI_Codes.Add("0x10", "telnet://");
      URI_Codes.Add("0x11", "imap:");
      URI_Codes.Add("0x12", "rtsp://");
      URI_Codes.Add("0x13", "urn:");
      URI_Codes.Add("0x14", "pop:");
      URI_Codes.Add("0x15", "sip:");
      URI_Codes.Add("0x16", "sips:");
      URI_Codes.Add("0x17", "tftp:");
      URI_Codes.Add("0x18", "btspp://");
      URI_Codes.Add("0x19", "btl2cap://");
      URI_Codes.Add("0x1A", "btgoep://");
      URI_Codes.Add("0x1B", "tcpobex://");
      URI_Codes.Add("0x1C", "irdaobex://");
      URI_Codes.Add("0x1D", "file://");
      URI_Codes.Add("0x1E", "urn:epc:id:");
      URI_Codes.Add("0x1F", "urn:epc:tag:");
      URI_Codes.Add("0x20", "urn:epc:pat:");
      URI_Codes.Add("0x21", "urn:epc:raw:");
      URI_Codes.Add("0x22", "urn:epc:");
      URI_Codes.Add("0x23", "urn:nfc:");
      URI_Codes.Add("0xFE", "http://nfc.springcard.com/?");
      foreach (string s in URI_Codes.AllKeys)
      {
        comboPrefixSp.Items.Add(URI_Codes[s]);
        cBPrefixURI.Items.Add(URI_Codes[s]);
      }
      comboPrefixSp.SelectedItem = "http://www.";
    }

    private string URI_Identifier_Code(byte code)
    {
      return URI_Codes[byteToStringHex(code)];
    }

    private byte URI_Identifier_Code(string code)
    {
      byte value = 0x00;
      string value_str = "";
      foreach (string s in URI_Codes.AllKeys)
      {
        if (URI_Codes[s] == code)
          value_str = s;
      }
      if (value_str != "")
        value = StringToByte(value_str);
      return value;
    }

    private byte[] decodeNDEF(byte[] ndef)
    {
      int rest_index = 0;
      if (ndef.Length > 3)
      {
        int payload_length = 4;

        if ((ndef[0] & SHORT_RECORD_FLAG) != 0)
        {
          payload_length = 1;
          payload_size = (int)ndef[PAYLOAD_LENGTH_INDEX];
        }
        else
        {
          payload_size = (int)(
            ndef[PAYLOAD_LENGTH_INDEX] << 24
            + ndef[PAYLOAD_LENGTH_INDEX + 1] << 16
            + ndef[PAYLOAD_LENGTH_INDEX + 2] << 8
            + ndef[PAYLOAD_LENGTH_INDEX + 3]);
        }
        int type_length = (int)ndef[TYPE_LENGTH_INDEX];
        int type_index = PAYLOAD_LENGTH_INDEX + payload_length;
        int payload_index = type_index + type_length;
        if ((ndef[0] & ID_LENGTH_FLAG) != 0)
        {
          type_index++;
          payload_index += (int)ndef[PAYLOAD_LENGTH_INDEX + payload_length];
        }
        byte tnf = (byte)(ndef[0] & 0x07);
        type_ndef_str = decodeTNF(tnf);
        byte[] type_array = new byte[type_length];
        for (int cpt = 0; cpt < type_length; cpt++)
        {
          if (cpt + type_index < ndef.Length) 
            type_array[cpt] = ndef[cpt + type_index];
        }
        record_name_str = createAsciiStrFromByteArray(type_array);
        rest_index = payload_index;
        if (payload_size > 0)
        {
          /* store payload in an array */
          payload_array = new byte[payload_size];
          for (int cpt = 0; cpt < payload_size; cpt++)
          {
            if (cpt + payload_index < ndef.Length)
              payload_array[cpt] = ndef[cpt + payload_index];
          }
          if (record_name_str == "Sp")
          {
            decodePayloadSp(payload_array);
            type_NDEF = TYPE_SP;
            tabChoiceSP.SelectedTab = tabPageSP;
          }
          if (record_name_str == "U")
          {
            decodePayloadURI(payload_array);
            if (type_NDEF == 0)
            {
              type_NDEF = TYPE_U;
              tabChoiceSP.SelectedTab = tabPageU;
            }
          }
          if (record_name_str == "T")
          {
            decodePayloadText(payload_array);
            if (type_NDEF == 0)
            {
              type_NDEF = TYPE_T;
              tabChoiceSP.SelectedTab = tabPageT;
            }
          }

          /*payload_str = URI_Identifier_Code(ndef[payload_index]) +
            createAsciiStrFromByteArray(payload_array);*/
          rest_index = payload_index + payload_size;
        }
      }
      int rest_size = ndef.Length - rest_index;
      byte[] rest_array = new byte[0];
      if (rest_size > 0)
      {
        rest_array = new byte[rest_size];
        for (int cpt = 0; cpt < rest_size; cpt++)
        {
          if (cpt + rest_index < ndef.Length)
            rest_array[cpt] = ndef[cpt + rest_index];
        }
      }
      return rest_array;
    }

    private byte[] encode_URI_Payload(byte URI_identifier_code, byte[] URI)
    {

      byte[] payloadEncoded = new byte[URI.Length + 1];
      payloadEncoded[0] = URI_identifier_code;
      if (URI.Length > 0)
      {
        for (int cpt = 0; cpt < URI.Length; cpt++)
          payloadEncoded[cpt + 1] = URI[cpt];
      }
      return payloadEncoded;
    }

    private byte[] encode_URI_NDEF(byte header,byte[] payload)
    {
      byte[] NDEFEncoded = new byte[256];

      if (payload.Length < 256)
      {
        // Short record
        NDEFEncoded = new byte[payload.Length + 4];
        NDEFEncoded[0] = header;
        NDEFEncoded[1] = 0x01;
        NDEFEncoded[2] = (byte)(payload.Length);
        NDEFEncoded[3] = (byte)'U';
        for (int cpt = 0; cpt < payload.Length; cpt++)
          NDEFEncoded[cpt + 4] = payload[cpt];
      }

      return NDEFEncoded;
    }

    private byte[] encode_T_NDEF(byte header,byte[] payload)
    {
      byte[] NDEFEncoded = new byte[256];

      if (payload.Length < 256)
      {
        // Short record
        NDEFEncoded = new byte[payload.Length + 4];
        NDEFEncoded[0] = header;
        NDEFEncoded[1] = 0x01;
        NDEFEncoded[2] = (byte)(payload.Length);
        NDEFEncoded[3] = (byte)'T'; 
        for (int cpt = 0; cpt < payload.Length; cpt++)
          NDEFEncoded[cpt + 4] = payload[cpt];
      }

      return NDEFEncoded;
    }

    private byte[] encode_T_Payload(string language, string title)
    {
      byte[] payloadEncoded;
      byte[] language_array = createByteArrayFromAsciiStr(language);
      byte[] title_array = createByteArrayFromAsciiStr(title);
      payloadEncoded = new byte[language_array.Length + title_array.Length + 1];
      payloadEncoded[0] = (byte) language_array.Length;
      int offset_payload = 1;
      for (int cpt = 0; cpt < language_array.Length; cpt++)
        payloadEncoded[cpt + offset_payload] = language_array[cpt];
      offset_payload += language_array.Length;
      for (int cpt =0;cpt<title_array.Length;cpt++)
        payloadEncoded[cpt + offset_payload] = title_array[cpt];

      return payloadEncoded;
    }

    
    private byte[] encode_Sp_NDEF(byte[] payload)
    {
      byte[] NDEFEncoded;
      //if (payload.Length < 255)
     //{
        NDEFEncoded = new byte[payload.Length + 5];
        NDEFEncoded[0] = SHORT_RECORD_HEADER;
        NDEFEncoded[1] = 0x02;
        NDEFEncoded[2] = (byte)payload.Length;
        NDEFEncoded[3] = (byte)'S';
        NDEFEncoded[4] = (byte)'p';
        int offset_payload = 5;
        for (int cpt = 0; cpt < payload.Length; cpt++)
          NDEFEncoded[cpt + offset_payload] = payload[cpt]; 
     // }
      return NDEFEncoded;
    }



  #region decode Payload

    const int STATUS_INDEX = 0;
    const int LANGUAGE_CODE_INDEX = 1;

    private void decodePayloadText(byte[] payload)
    {
      Card_Title_present = true;
      /* first, decode status byte : retreive language code length */
      int langage_code_size = (int)(payload[STATUS_INDEX] & 0x3F);

      /* retreive language code */
      byte[] language_array = new byte[langage_code_size];
      for (int i = 0; i < langage_code_size; i++)
      {
        if (LANGUAGE_CODE_INDEX + i < payload.Length)
          language_array[i] = payload[LANGUAGE_CODE_INDEX + i];
      }
      Card_Title_lang = createAsciiStrFromByteArray(language_array);

      /* retreive text*/
      int text_index = LANGUAGE_CODE_INDEX + langage_code_size;
      if (text_index < payload.Length)
      {
        byte[] text_array = new byte[payload.Length - text_index];
        for (int i = 0; i < text_array.Length; i++)
        {
          if (i + text_index < payload.Length)
            text_array[i] = payload[i + text_index];
        }

        Card_Title_content = createAsciiStrFromByteArray(text_array);
        
      }


    }

    private void decodePayloadURI(byte[] payload)
    {
      if (payload.Length > 1)
      {
        byte [] payload_array_tmp = new byte[payload.Length - 1];
        for (int cpt = 0; cpt < payload_array_tmp.Length; cpt++)
        {
          payload_array_tmp[cpt] = payload[cpt + 1];
        }
        Card_URI_prefix = URI_Identifier_Code(payload[0]);
        Card_URI_content = createAsciiStrFromByteArray(payload_array_tmp);
        Card_URI_Tot = Card_URI_prefix + Card_URI_content;
        payload_str = URI_Identifier_Code(payload[0]) +
          createAsciiStrFromByteArray(payload_array_tmp);
      }

    }


    private void decodePayloadSp(byte[] payload)
    {
      byte[] rest;
      do
      {
        rest = decodeNDEF(payload);
        payload = rest;
      } while (rest.Length > 0);
    }

    #endregion

#endregion

#region PC/SC function

    string lastApduIn = "";
    string lastApduOut = "";
    int hCard =0;

    public APDURsp ConnectAndTransmit(APDUCmd ApduIn)
    {

      SPCSC.ConnectCard();
      //hCard = SPCSC.GetHCard();
     // AttachLibrary(hCard);
      APDURsp Resp = Transmit(ApduIn);
     // DetachLibrary(hCard);
      SPCSC.Disconnect();


      return Resp;
    }

    public APDURsp Transmit(APDUCmd ApduIn)
    {
      APDURsp Resp = new APDURsp();
      SPCSC.Transmit(ApduIn, ref Resp);
      // InfoText += "Sending " + ApduIn.APDUStr + "\n"; 
      lastApduIn = ApduIn.APDUStr;
      lastApduOut = Resp.APDUStr + " - " + SPCSC.StatusWordsToStr(Resp.SW1, Resp.SW2);
      //InfoText += "Received " + Resp.APDUStr + " - " + SPCSC.StatusWordsToStr(Resp.SW1, Resp.SW2) + "\n";
      if ((Resp.SW1 != 0x90) || (Resp.SW2 != 0x00))
      {
        string ErrorText = "Error while sending " + ApduIn.APDUStr
            + " : " + Resp.APDUStr + " - " + SPCSC.StatusWordsToStr(Resp.SW1, Resp.SW2);
        //MessageBox.Show(ErrorText);
      }
      return Resp;
    }
#endregion

#region open URL
   // [System.Runtime.InteropServices.DllImportAttribute("shell32.dll", SetLastError = true)]
   // static extern int ShellExecute(int hwnd, string lpOperation, string lpFile, string lpParameters, string lpDirectory, int nShowCmd);
    //
    public void RunUrl(string sUrl, bool bMaximized)
    {
      OpenUrlFrm openUrl = new OpenUrlFrm(sUrl, bMaximized);
      openUrl.ReadRegister();

      if (!openUrl.alwaysOpen)
        openUrl.ShowDialog();
      else
        openUrl.RunUrl(sUrl, bMaximized);
      /*if (MessageBox.Show("Would you like to open "+sUrl+"?",
      "Opening "+sUrl,
      MessageBoxButtons.OKCancel) == DialogResult.OK)
        ShellExecute(0, "open", sUrl, string.Empty, string.Empty, (bMaximized == true) ? 3 : 1);*/
    }


#endregion

#region write card

    private void WriteCard(byte[] data)
    {
      butWrite.Enabled = false;
      if ((card_type == ULTRALIGHT) || (card_type == ULTRALIGHTC))
        WriteUL(data);
      if (card_type == JEWEL)
        WriteJewel(data);
      if (card_type == MIFARE_1K || card_type == MIFARE_4K)
        WriteMifare1K(data);
      butWrite.Enabled = true;
    }

    #region Mifare UL

    public void WriteUL(byte[] DataIn)
    {
      byte card_block = 0x04;
      byte[] data = new byte[16];
     /* SPCSC.ConnectCard();
      hCard = SPCSC.GetHCard();
      AttachLibrary(hCard);*/

      // write NFC header
      byte[] NFC_H = {0xE1,0x10,0x06,0x00};
      APDUCmd writeHeader = new APDUCmd(0xFF, 0xD6, 0x00, 0x03, NFC_H);
      APDURsp Rsp = ConnectAndTransmit(writeHeader);
     // MifUlC_WriteBlock(0x03, NFC_H);
      for (int offset = 0; offset < DataIn.Length; offset = offset + 16)
      {
        data[0] = DataIn[offset];
        for (int cpt = 1; cpt < 16; cpt++)
        {
          if (offset + cpt < DataIn.Length)
            data[cpt] = DataIn[offset +cpt];
          else
            data[cpt] = 0x00;
        }

        
        APDUCmd writeUl = new APDUCmd(0xFF, 0xD6, 0x00, card_block, data);
       // MessageBox.Show(writeUl.APDUStr);
        Rsp = ConnectAndTransmit(writeUl);
        //MifUlC_WriteBlock(card_block, data);

        card_block += 0x04;
      }
      
      /*DetachLibrary(hCard);
      SPCSC.Disconnect();*/
    }


    public void WriteJewel(byte[] DataIn)
    {
      byte card_block = 0x08;
      byte[] data = new byte[1];
    
      // NDEF header
      data[0] = 0xE1;
      APDUCmd writeUl = new APDUCmd(0xFF, 0xD6, 0x00, card_block, data);
      APDURsp Rsp = ConnectAndTransmit(writeUl);
      card_block += 0x01;

      data[0] = 0x10;
      writeUl = new APDUCmd(0xFF, 0xD6, 0x00, card_block, data);
      Rsp = ConnectAndTransmit(writeUl);
      card_block += 0x01;

      data[0] = 0xE0;
      writeUl = new APDUCmd(0xFF, 0xD6, 0x00, card_block, data);
      Rsp = ConnectAndTransmit(writeUl);
      card_block += 0x01;

      data[0] = 0x00;
      writeUl = new APDUCmd(0xFF, 0xD6, 0x00, card_block, data);
      Rsp = ConnectAndTransmit(writeUl);
      card_block += 0x01;

      for (int cpt = 0; cpt < DataIn.Length; cpt++)
      {
        data[0] = DataIn[cpt];
        writeUl = new APDUCmd(0xFF, 0xD6, 0x00, card_block, data);
        Rsp = ConnectAndTransmit(writeUl);
        card_block += 0x01;
      }

    }

    #endregion

    #region Mifare1K

    private const int SZ_SECTOR_MIF1K = 48;
    private const byte INFO_BYTE = 0x00;
    private const byte AID_1 = 0x03;
    private const byte AID_2 = 0xE1;
    private byte[] MIFARE_KEY_A = { 0xA0, 0xA1, 0xA2, 0xA3, 0xA4, 0xA5 };
    private byte[] MIFARE_KEY_NFC = { 0xD3, 0xF7, 0xD3, 0xF7, 0xD3, 0xF7};
    private byte[] MIFARE_WRITE_KEY = { 0x53, 0x43, 0x4E, 0x46, 0x43, 0x42 };
    private byte[] ACCESS_MAD = {0xA0,0xA1,0xA2, 0xA3,0xA4, 0xA5,0x78,0x77,0x88,0xC1 };
    private byte[] ACCESS_NFC = {0xD3,0xF7, 0xD3, 0xF7, 0xD3, 0xF7, 0x7F,0x07,0x88,0x40};

    public void WriteMifare1K(byte [] data)
    {
      /* step 1 : how many sectors? */
      int nb_sectors = (int) (data.Length / SZ_SECTOR_MIF1K);

      if (data.Length % SZ_SECTOR_MIF1K > 0)
        nb_sectors++;

      /* step 2 : set MAD sector array */
      byte[] mad_array = new byte[32];
      for (int i = 0; i < 32; i++)
        mad_array[i] = 0;
      int offset_mad = 2;
      int offset_crc = 0;
      byte[] crc_array = new byte[30];
      for (int i = 0; i < 30; i++)
        crc_array[i] = 0;

      for (int i = 0; i < nb_sectors; i++)
      {
        crc_array[offset_crc] = AID_1;
        mad_array[offset_mad] = AID_1;
        offset_crc++;
        offset_mad++;
        crc_array[offset_crc] = AID_2;
        mad_array[offset_mad] = AID_2;
        offset_crc++;
        offset_mad++;
      }
     

      /* step 3 : calculate CRC, to put in MAD array */
      mad_array[0] = CalcMifCrc(crc_array);
      /* step 4 : write MAD sector */
      WriteDataMifare1K(1, mad_array);
      /* step 5 : change MAD sector read key + Access conditions */
      byte[] trailor_array = new byte[16];
      for (int i = 0; i < 10; i++)
        trailor_array[i] = ACCESS_MAD[i];
      for (int i = 10; i < 16; i++)
        trailor_array[i] = MIFARE_WRITE_KEY[i-10];
      WriteDataMifare1K(3, trailor_array);

      int offset_data = 0;
      /* for each sector to be written */
      for (int sector = 0; sector < nb_sectors; sector++)
      {

        byte[] dataToWrite = new byte[SZ_SECTOR_MIF1K];

        for (int i = 0; i < SZ_SECTOR_MIF1K; i++)
        {
          if (i < data.Length - offset_data)
            dataToWrite[i] = data[i + offset_data];
          else
            dataToWrite[i] = 0;
        }

        offset_data += SZ_SECTOR_MIF1K;
        
        byte blockToWrite =  (byte)((sector+1) * 4);
        /* write sector */
        WriteDataMifare1K(blockToWrite, dataToWrite);

        /* change sector read key + Access conditions */
        for (int i = 0; i < 10; i++)
          trailor_array[i] = ACCESS_NFC[i];
        for (int i = 10; i < 16; i++)
          trailor_array[i] = MIFARE_WRITE_KEY[i-10];
        WriteDataMifare1K( (byte) (blockToWrite+0x03), trailor_array);
      }
    }


    public void WriteDataMifare1K(byte block, byte[] data)
    {
      APDUCmd WriteApdu = new APDUCmd(0xFF, 0xF4, (byte)((block & 0xFF00) >> 8),
        (byte)((block) & 0x00FF), data);
      APDURsp Resp = ConnectAndTransmit(WriteApdu);
    }

    public byte CalcMifCrc(byte [] array_in) 
    {
      byte crc,i,j;

      crc = 0xE3;

      for(j =0;j<array_in.Length;j++)
      {
        crc = (byte) (crc ^ array_in[j]);
        for(i=0;i<8;i++)
        {
          if((crc&0x80)>0)
            crc = (byte) ((crc << 1) ^ 0x1D);
          else
            crc = (byte) (crc << 1);
        }
      }
      return crc;
    }



    #endregion


#endregion

#region Mifare Ultralight C


    #region DLL Imports
    // Attach Library : use this function before using any function of the DLL
    [DllImport("pcsc_mifulc.dll", EntryPoint = "SCardMifUlC_AttachLibrary", SetLastError = true)]
    internal static extern uint AttachLibrary(int hCard);

    // Detach Library : use this function after using functions of the DLL
    [DllImport("pcsc_mifulc.dll", EntryPoint = "SCardMifUlC_DetachLibrary", SetLastError = true)]
    internal static extern uint DetachLibrary(int hCard);

    // Authenticate : use this function to authenticate with a 16-byte key
    [DllImport("pcsc_mifulc.dll", EntryPoint = "SCardMifUlC_Authenticate", SetLastError = true)]
    internal static extern uint Authenticate(int hCard, byte[] key);

    // Read : read 4 blocks in a row (16 bytes)
    [DllImport("pcsc_mifulc.dll", EntryPoint = "SCardMifUlC_Read", SetLastError = true)]
    internal static extern uint Read(int hCard, byte page, [In, Out] byte[] data);

    // Write : write 4 blocks in a row (16 bytes)
    [DllImport("pcsc_mifulc.dll", EntryPoint = "SCardMifUlC_Write", SetLastError = true)]
    internal static extern uint Write(int hCard, byte page, byte[] data);

    // Write4 : write one block (4 bytes)
    [DllImport("pcsc_mifulc.dll", EntryPoint = "SCardMifUlC_Write4", SetLastError = true)]
    internal static extern uint Write4(int hCard, byte page, byte[] data);

    #endregion

    public byte[] BLANK_KEY =
new byte[16] { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
      0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

    public byte[] TRANSPORT_KEY =
new byte[16] { 0x49, 0x45, 0x4D, 0x4B, 0x41, 0x45, 0x52, 0x42,
      0x21, 0x4E, 0x41, 0x43, 0x55, 0x4F, 0x59, 0x46 };


    #region Mifare UltralightC functions : Authenticate, Read, Write
    private void MifUlC_Authenticate()
    {
      // let's try to authenticate with our keys
      uint rc = Authenticate(hCard, BLANK_KEY);
      if (rc != 0)
        rc = Authenticate(hCard, TRANSPORT_KEY);
    }

    private void MifUlC_ReadFullCard()
    {
      byte page;
      byte[] data = new byte[16];
      uint rc;
      int offset_cardContent = 0;
      card_content = new byte[192];
      SPCSC.ConnectCard();
      hCard = SPCSC.GetHCard();
      AttachLibrary(hCard);
      card_type = ULTRALIGHT;
      for (page =4; page < 44; page += 4)
      {
        MifUlC_Authenticate(); 
        rc = Read(hCard, page, data);
        for (int i = 0; i < 16; i++)
        {
          if (offset_cardContent + i < card_content.Length)
            card_content[offset_cardContent + i] = data[i];
        }
        offset_cardContent += 16;
        if ((page >= 16) && (rc == 0))
          card_type = ULTRALIGHTC;
      }

      DetachLibrary(hCard);
      SPCSC.Disconnect();

    }

    // write 4 blocks (4x4 bytes)
    private void MifUlC_WriteBlock(byte index, byte[] block_content)
    {
      uint rc;
      if (block_content.Length == 16)
      {
        MifUlC_Authenticate();
        rc = Write(hCard, index, block_content);
      }
    }

    #endregion

#endregion


    private void pBClose_Click(object sender, EventArgs e)
    {
      this.Close();
      this.Dispose();
    }

    private void rTBURI_Click(object sender, EventArgs e)
    {
      RunUrl(payload_str, true);
    }

    private void ButWrite_Click(object sender, EventArgs e)
    {
      if (rTBNDEFEnc.Text.Length >0)
      {
        byte[] Data = encodeTLV(NDEF_MESSAGE_TLV, createByteArrayFromStr(rTBNDEFEnc.Text));
        WriteCard(Data);
      }
    }

    private void PosterFrm_Shown(object sender, EventArgs e)
    {
      /* Showing configuration form - User has to select reader */
      ChooseReaderFrm cf = new ChooseReaderFrm(this, true);
      cf.ReadRegister();

      if ((cf.ReaderName == "") || (!cf.connect))
        cf.ShowDialog();
      else
        SetReader(cf.ReaderName);
    }

    private void butEnTitle_Click(object sender, EventArgs e)
    {
      byte[] TPayload_result = encode_T_Payload(cBLangText.Text, rTBText.Text);
      byte[] TNDEF_result = encode_T_NDEF(SHORT_RECORD_HEADER,TPayload_result);
      rTBNDEFEnc.Text = createStrFromByteArray(TNDEF_result);
    }

    private void butEncURI_Click(object sender, EventArgs e)
    {
      byte[] UPayload_result = encode_URI_Payload(URI_Identifier_Code(cBPrefixURI.Text),
        createByteArrayFromAsciiStr(rTBURIContent.Text));
      byte[] UNDEF_result = encode_URI_NDEF(SHORT_RECORD_HEADER, UPayload_result);
      rTBNDEFEnc.Text = createStrFromByteArray(UNDEF_result);
    }

    private void butEncSp_Click(object sender, EventArgs e)
    {
      if (cBTitle.Checked)
      {
        // Sp with title
        // encode title
        byte[] TPayload_result = encode_T_Payload(cBLangTitleSp.Text, tBtitleSp.Text);
        byte[] TNDEF_result = encode_T_NDEF(SR_ME_HEADER, TPayload_result);
        // encode URI
        byte[] UPayload_result = encode_URI_Payload(URI_Identifier_Code(comboPrefixSp.Text),
       createByteArrayFromAsciiStr(rTBURISp.Text));
        byte[] UNDEF_result = encode_URI_NDEF(SR_MB_HEADER, UPayload_result);
        // concat tables
        byte[] SPPayload = new byte[TNDEF_result.Length + UNDEF_result.Length];
        TNDEF_result.CopyTo(SPPayload, 0);
        UNDEF_result.CopyTo(SPPayload, TNDEF_result.Length);
        byte[] SNEDF = encode_Sp_NDEF(SPPayload);
        rTBNDEFEnc.Text = createStrFromByteArray(SNEDF);
      }
      else
      {
        // Sp without title
        byte[] UPayload_result = encode_URI_Payload(URI_Identifier_Code(comboPrefixSp.Text),
          createByteArrayFromAsciiStr(rTBURISp.Text));
        byte[] UNDEF_result = encode_URI_NDEF(SHORT_RECORD_HEADER, UPayload_result);
        byte[] SNEDF = encode_Sp_NDEF(UNDEF_result);
        rTBNDEFEnc.Text = createStrFromByteArray(SNEDF);
      }

    }

    private void butCopy_Click(object sender, EventArgs e)
    {
      Clipboard.SetText(rTBNDEFEnc.Text);
    }

    private void rTBNDEFEnc_TextChanged(object sender, EventArgs e)
    {
      byte[] NDEFArray = createByteArrayFromStr(rTBNDEFEnc.Text);
      lblSize.Text = "NDEF Size : " + NDEFArray.Length;
      bool warningSize = false;
      if (card_type == ULTRALIGHT)
      {
        lblSize.Text += "/" + MAX_NDEF_UL;
        if (NDEFArray.Length > MAX_NDEF_UL)
          warningSize = true;
      }
      if (card_type == ULTRALIGHTC)
      {
        lblSize.Text += "/" + MAX_NDEF_ULC;
        if (NDEFArray.Length > MAX_NDEF_ULC)
          warningSize = true;
      }
      if (card_type == JEWEL)
      {
        lblSize.Text += "/" + MAX_NDEF_JEWEL;
        if (NDEFArray.Length > MAX_NDEF_JEWEL)
          warningSize = true;
      }
      if (warningSize)
      {
        rTBAdvise.Text = "NDEF size too big";
        butWrite.Enabled = false;
      }
      else
      {
        butWrite.Enabled = true;
      }
      
      
    }



  }
}